import sys
import typing as t

from click import Option, Parameter
from uvicorn.config import Config
from uvicorn.server import Server
from uvicorn.supervisors import ChangeReload, Multiprocess

from unfazed.command import BaseCommand


class Command(BaseCommand):
    help_text = "Run a Uvicorn server"

    def add_arguments(self) -> t.Sequence[Parameter]:
        return [
            Option(
                ["--host"],
                type=str,
                help="The host to listen on",
                default="127.0.0.1",
                show_default=True,
            ),
            Option(
                ["--port"],
                type=int,
                help="The port to listen on",
                default=9527,
                show_default=True,
            ),
            Option(
                ["--reload"],
                is_flag=True,
                help="Enable auto-reload",
                default=True,
                show_default=True,
            ),
            Option(
                ["--workers"],
                type=int,
                help="Number of worker processes",
                default=1,
                show_default=True,
            ),
            Option(
                ["--log-level"],
                type=str,
                help="Log level",
                default="info",
                show_default=True,
            ),
        ]

    async def handle(
        self,
        host: str,
        port: str,
        reload: bool,
        workers: int,
        log_level: str,
    ) -> None:
        config = Config(
            self.unfazed,
            host=host,
            port=port,
            reload=reload,
            workers=workers,
            log_level=log_level,
        )
        server = Server(config)

        sock = config.bind_socket()
        if config.should_reload:
            ChangeReload(config, target=server.run, sockets=[sock]).run()

        elif config.workers > 1:
            Multiprocess(config, target=server.run, sockets=[sock]).run()

        else:
            await server.serve([sock])

        if not server.started and not config.should_reload and config.workers == 1:
            sys.exit(3)



import typing as t
import warnings

from starlette.routing import BaseRoute, Mount, Route, WebSocketRoute


def path(
    path: str,
    endpoint: t.Callable | t.Sequence[Route] | t.Sequence[WebSocketRoute],
    *,
    methods: t.Sequence[str] = None,
    name: str = None,
    websocket: bool = False,
) -> BaseRoute:
    if isinstance(endpoint, t.Callable):
        if websocket:
            return WebSocketRoute(path, endpoint, methods=methods, name=name)
        else:
            return Route(path, endpoint, methods=methods, name=name)

    if methods is not None:
        warnings.warn("methods will be ignored when endpoint is a list of routes")

    return Mount(path, routes=endpoint, name=name)



from .app import AppCommand, AppConfig
from .command import BaseCommandMethod
from .middleware import MiddleWareProtocol
from .orm import OrmDataBase, OrmMeta, OrmModel, OrmQuerySet

__all__ = [
    "OrmDataBase",
    "OrmMeta",
    "OrmModel",
    "OrmQuerySet",
    "AppConfig",
    "AppCommand",
    "BaseCommandMethod",
    "MiddleWareProtocol",
]



import importlib
import os
import typing as t

from pydantic import BaseModel
from unfazed.const import UNFAZED_SETTINGS_MODULE
from unfazed.types import InstalledApp, Middlewares


def get_settings_module() -> t.Mapping:
    settings_module = os.environ.get(UNFAZED_SETTINGS_MODULE)
    if not settings_module:
        raise Exception("UNFAZED_SETTINGS_MODULE environment variable is not set")
    try:
        settingskv = importlib.import_module(settings_module).__dict__
    except ImportError:
        raise ImportError(f"Could not import settings module {settings_module}")

    return settingskv


class BaseSettings(BaseModel):
    INSTALLED_APPS: InstalledApp
    MIDDLEWARE: Middlewares
    DEBUG: bool
    ROOT_URLCONF: str
    DATABASE: t.Mapping | None = None
    PROJECT_NAME: str

    @classmethod
    def setup(cls):
        # get all key-value settings from the settings module
        return cls(**get_settings_module())


def settings_from_key(key: str) -> t.Mapping:
    settings_kv = get_settings_module()

    if key not in settings_kv:
        raise KeyError(f"{key} is not found in settings module")

    return settings_kv[key]



"""
run module for {{project}}
"""

import os
import sys
import warnings


def main() -> None:
    os.environ.setdefault("UNFAZED_SETTINGS_MODULE", "{{project}}.settings")
    try:
        from unfazed.core import Unfazed

        unfazed = Unfazed()
        unfazed.setup()
    except ImportError:
        warnings.warn("Could not import unfazed. Did you forget to install it?")
        sys.exit(1)

    except Exception as e:
        warnings.warn(f"An error occurred: {e}")
        sys.exit(1)

    unfazed.execute_command_from_argv()


if __name__ == "__main__":
    main()



import typing as t

if t.TYPE_CHECKING:
    from unfazed.core import Unfazed

from unfazed.protocol import OrmModel


class ModelCenter:
    def __init__(
        self,
        unfazed: "Unfazed",
        models: t.List[t.Type[OrmModel]] | None = None,
    ) -> None:
        self.unfazed = unfazed
        self.models = models

    def add_model(self, model: t.Type[OrmModel]) -> None:
        if self.models is None:
            self.models = []
        self.models.append(model)



import typing as t
from importlib import import_module
from threading import Lock

from starlette.applications import Starlette
from starlette.middleware import Middleware
from starlette.routing import BaseRoute
from starlette.types import ASGIApp
from unfazed.app.registry import AppCenter
from unfazed.command import CommandCenter
from unfazed.conf import BaseSettings
from unfazed.middleware.registry import MiddleWareCenter
from unfazed.orm.registry import ModelCenter


class Unfazed(Starlette):
    version = "0.0.1"

    def __init__(
        self,
        debug: bool = False,
        routes: t.Sequence[BaseRoute] | None = None,
        middleware: t.Sequence[Middleware] | None = None,
    ) -> None:
        self._ready = False
        self._settings: BaseSettings = None
        self._app_center: AppCenter = None
        self._command_center: CommandCenter = None
        self._model_center: ModelCenter = None
        super().__init__(debug=debug, routes=routes, middleware=middleware)

    @property
    def settings(self) -> BaseSettings:
        if not self._settings:
            self.setup_settings()

        return self._settings

    @property
    def app_center(self) -> AppCenter:
        if not self._app_center:
            self.setup_app()

        return self._app_center

    @property
    def command_center(self) -> CommandCenter:
        if not self._command_center:
            self.setup_commands()

        return self._command_center

    @property
    def model_center(self) -> ModelCenter:
        if not self._model_center:
            self.setup_models()

        return self._model_center

    def setup_settings(self) -> None:
        self._settings = BaseSettings.setup()

    def setup_app(self):
        self._app_center = AppCenter(self, self.settings.INSTALLED_APPS)
        self._app_center.setup()

    def setup_commands(self):
        self._command_center = CommandCenter(
            unfazed=self,
            name=self.settings.PROJECT_NAME,
            commands=self.app_center.commands,
        )
        self._command_center.setup()

    def setup_models(self):
        # self._model_center = ModelCenter(
        #     unfazed=self,
        #     models=self.app_center.models,
        # )

        # TODO
        # wait for orm to be implemented
        pass

    def setup_middlewares(self):
        # load middlewares
        middleware_center = MiddleWareCenter(self.settings.MIDDLEWARE)
        middleware_center.setup()

        for m in middleware_center.middlewares:
            self.add_middleware(m)

    def setup_routes(self):
        # add routes from settings.ROOT_URLCONF
        root_url_module = import_module(self.settings.ROOT_URLCONF)

        if not hasattr(root_url_module, "urlpatterns"):
            raise ValueError("ROOT_URLCONF should have urlpatterns defined")

        for route in root_url_module.urlpatterns:
            # TODO
            # check if route is a valid route
            # check if route app is installed
            self.router.routes.append(route)

    def setup(self) -> None:
        """
        steps to setup the application
        """

        if self._ready:
            return
        with Lock():
            if self._ready:
                return

            self.setup_settings()
            self.setup_app()

            self.setup_commands()
            self.setup_models()

            self.setup_middlewares()
            self.setup_routes()

            self._ready = True

    def setup_cli(self):
        self._command_center = CommandCenter(unfazed=self)
        self._command_center.setup_cli()

    def execute_command_from_argv(self):
        self._command_center.main()

    def build_middleware_stack(self) -> ASGIApp:
        # debug = self.debug
        # error_handler = None
        # exception_handlers: dict[
        #     t.Any, t.Callable[[Request, Exception], Response]
        # ] = {}

        # for key, value in self.exception_handlers.items():
        #     if key in (500, Exception):
        #         error_handler = value
        #     else:
        #         exception_handlers[key] = value

        # middleware = (
        #     [Middleware(ServerErrorMiddleware, handler=error_handler, debug=debug)]
        #     + self.user_middleware
        #     + [
        #         Middleware(
        #             ExceptionMiddleware, handlers=exception_handlers, debug=debug
        #         )
        #     ]
        # )

        middleware = self.user_middleware

        app = self.router
        for cls, args, kwargs in reversed(middleware):
            app = cls(app=app, *args, **kwargs)
        return app



import typing as t

from starlette.types import ASGIApp, Receive, Scope, Send


class MiddleWareProtocol(t.Protocol):
    def __init__(self, app: ASGIApp) -> None: ...
    async def __call__(self, scope: Scope, receive: Receive, send: Send): ...



import importlib
import typing as t
import warnings
from pathlib import Path

from unfazed.command import BaseCommand
from unfazed.protocol import OrmModel
from unfazed.utils.module_loading import import_string

if t.TYPE_CHECKING:
    from unfazed.core import Unfazed


class AppConfig:
    # name of the app
    # also the import path of the app
    name: str

    def __init__(self, unfazed: "Unfazed") -> None:
        self._ready = False
        self.commands = []
        self.models = []
        self.unfazed = unfazed

    @property
    def app_module(self) -> str:
        return self.__module__.rsplit(".", 1)[0]

    @property
    def app_path(self) -> Path:
        path = importlib.import_module(self.app_module).__path__._path[0]

        return Path(path)

    def setup(self) -> None:
        if self._ready:
            return

        # collect commands
        # list all files in the app/commands
        commands_path = self.app_path.joinpath("commands")

        if commands_path.exists() and commands_path.is_dir():
            command_file_list = commands_path.glob("*.py")

            commands = []
            for command_file in command_file_list:
                # check if BaseCommand is in the file
                path = f"{self.app_module}.commands.{command_file.stem}.Command"
                try:
                    command_cls = import_string(path)
                    if not issubclass(command_cls, BaseCommand):
                        raise ImportError("Not a valid command")

                except ImportError as e:
                    warnings.warn(
                        f"Command {command_file.stem} is not a valid command: err {e}"
                    )
                    continue

                commands.append(
                    command_cls(
                        self.unfazed,
                        app_label=self.name,
                        name=command_file.stem,
                    )
                )

            self.commands = commands

        # collect models
        # import all models from models module
        # models module may be models.py or models/__init__.py
        # all models should inherit from unfazed.orm.models.Model
        models: t.List[t.Type[OrmModel]] = []
        models_dir_path = self.app_path.joinpath("models")
        models_file_path = self.app_path.joinpath("models.py")
        # breakpoint()
        if models_dir_path.exists() or models_file_path.exists():
            models_module = importlib.import_module(f"{self.app_module}.models")
            for name in dir(models_module):
                obj = getattr(models_module, name)
                if isinstance(obj, type) and hasattr(obj, "meta"):
                    # if isinstance(obj, type) and issubclass(obj, OrmModel):
                    if not obj.meta.abstract:
                        models.append(obj)

            self.models = models

        # TODO
        # collect url/admin

        # exec ready method
        self.ready()

        self._ready = True

        return None

    def ready(self):
        pass



from unfazed.protocol import OrmQuerySet


class QuerySet(OrmQuerySet):
    def __init__(self) -> None:
        self._using = "default"

    def using(self, alias: str):
        self._using = alias
        return self

    def filter(self, **kwargs):
        return self

    def exclude(self, **kwargs):
        return self

    async def all(self):
        return self

    async def get(self, **kwargs):
        return self

    async def create(self, **kwargs):
        return self

    async def update(self, **kwargs):
        return self

    async def delete(self, **kwargs):
        return self

    async def bulk_create(self, **kwargs):
        return self

    async def get_or_create(self, **kwargs):
        return self

    async def update_or_create(self, **kwargs):
        return self

    async def execute(self):
        return self



from pathlib import Path

import unfazed
from unfazed.command import BaseCommandWithTemplate


class Command(BaseCommandWithTemplate):
    help_text = "Create a new project with template"
    template = Path(unfazed.__path__[0], "template/project")



import typing as t

from click import Parameter


class BaseCommandMethod[T](t.Protocol):
    def add_arguments(self) -> t.Sequence[Parameter]: ...
    async def handle(self, **option: t.Optional[t.Any]) -> t.Optional[T]: ...



import typing as t
from threading import Lock

from unfazed.app.base import AppConfig
from unfazed.utils.module_loading import import_string

if t.TYPE_CHECKING:
    from unfazed.core import Unfazed


class AppCenter:
    def __init__(self, unfazed: "Unfazed", installed_apps: t.Sequence[str]) -> None:
        self.unfazed = unfazed
        self.installed_apps = installed_apps

        self._ready = False
        self._loading = False
        self._lock = Lock()
        self._app_list = []

        self.commands = []
        self.models = []
        # self.urls = []
        # self.admin_models = []

    def setup(self) -> None:
        if self._ready:
            return

        with self._lock:
            if self._ready:
                return

            if self._loading:
                raise RuntimeError("circular dependency detected")

            self._loading = True

            app_list: t.Sequence[AppConfig] = []
            for app_path in self.installed_apps:
                temp_app = self.load_app(app_path)

                self.register_commands(temp_app)
                self.register_models(temp_app)

                app_list.append(temp_app)

            self._app_list = app_list

            self._ready = True
            self._loading = False

        return None

    def load_app(self, app_path: str) -> AppConfig:
        if not app_path.endswith(".AppConfig"):
            app_path += ".app.AppConfig"
        app_cls: t.Type[AppConfig] = import_string(app_path)

        app: AppConfig = app_cls(self.unfazed)
        app.setup()

        return app

    def register_commands(self, app: AppConfig) -> None:
        self.commands.extend(app.commands)

    def register_models(self, app: AppConfig) -> None:
        # TODO
        # wait for orm registry to be implemented
        # validate there are no duplicate db_table
        # self.models.extend(app.models)
        pass



import typing as t

import sqlalchemy

# from pydantic.aliases import AliasChoices, AliasPath
from pydantic.fields import FieldInfo
from pydantic.types import Discriminator

type STE = sqlalchemy.types.TypeEngine


class ModelField[T](FieldInfo):
    def __init__(
        self,
        default: t.Any | None = None,
        *,
        default_factory: t.Callable[[], t.Any] | None = None,
        alias: str | None = None,
        alias_priority: int | None = None,
        # validation_alias: str | AliasPath | AliasChoices | None = None,
        serialization_alias: str | None = None,
        title: str | None = None,
        description: str | None = None,
        examples: list[t.Any] | None = None,
        exclude: bool | None = None,
        discriminator: str | Discriminator | None = None,
        frozen: bool | None = None,
        validate_default: bool | None = None,
        repr: bool | None = None,
        init: bool | None = None,
        init_var: bool | None = None,
        kw_only: bool | None = None,
        gt: float | None = None,
        ge: float | None = None,
        lt: float | None = None,
        le: float | None = None,
        multiple_of: float | None = None,
        strict: bool | None = None,
        min_length: int | None = None,
        max_length: int | None = None,
        db_column: str | None = None,
        db_comment: str | None = None,
        db_default: str | None = None,
        db_index: str | None = None,
        db_unique: bool | None = None,
        db_null: bool | None = None,
        primary_key: bool | None = None,
    ):
        super().__init__(
            default=default,
            default_factory=default_factory,
            alias=alias,
            alias_priority=alias_priority,
            # validation_alias=validation_alias,
            serialization_alias=serialization_alias,
            title=title,
            description=description,
            examples=examples,
            exclude=exclude,
            discriminator=discriminator,
            frozen=frozen,
            validate_default=validate_default,
            repr=repr,
            init=init,
            init_var=init_var,
            kw_only=kw_only,
            gt=gt,
            ge=ge,
            lt=lt,
            le=le,
            multiple_of=multiple_of,
            strict=strict,
            min_length=min_length,
            max_length=max_length,
        )

        self.db_column = db_column
        self.db_comment = db_comment or description
        self.db_default = db_default or self.get_default()
        self.db_index = db_index
        self.db_unique = db_unique
        self.db_null = db_null
        self.primary_key = primary_key

    def get_column(self, name: str) -> sqlalchemy.Column:
        return sqlalchemy.Column(
            name,
            self.get_column_type(),
            primary_key=self.primary_key,
            nullable=self.db_null,
            default=self.db_default,
            comment=self.db_comment,
            index=self.db_index,
            unique=self.db_unique,
        )

    def get_column_type(self) -> STE:
        raise NotImplementedError

    def to_python(self, value: T) -> T:
        return value

    def to_db(self, value: T) -> T:
        return value


class String(ModelField):
    def __init__(
        self,
        max_length: int,
        **kwargs,
    ):
        super().__init__(
            max_length=max_length,
            **kwargs,
        )

    def get_column_type(self) -> STE:
        return sqlalchemy.String(length=self.max_length)


class Text(ModelField):
    def get_column_type(self) -> STE:
        return sqlalchemy.Text()


class JsonText(ModelField):
    def get_column_type(self) -> STE:
        return sqlalchemy.Text()


class Integer(ModelField):
    def get_column_type(self) -> STE:
        return sqlalchemy.Integer()


class BigInteger(ModelField):
    def get_column_type(self) -> STE:
        return sqlalchemy.BigInteger()


class SmallInteger(ModelField):
    def get_column_type(self) -> STE:
        return sqlalchemy.SmallInteger()


class DateTime(ModelField):
    def get_column_type(self) -> STE:
        return sqlalchemy.DateTime()


class Date(ModelField):
    def get_column_type(self) -> STE:
        return sqlalchemy.Date()


class TsDateTime(ModelField):
    def get_column_type(self) -> STE:
        return sqlalchemy.Integer()


class TsDate(ModelField):
    def get_column_type(self) -> STE:
        return sqlalchemy.Integer()


class Time(ModelField):
    def get_column_type(self) -> STE:
        return sqlalchemy.Time()


class Decimal(ModelField):
    def get_column_type(self) -> STE:
        return sqlalchemy.DECIMAL()


class Float(ModelField):
    def get_column_type(self) -> STE:
        return sqlalchemy.Float()



from unfazed.protocol import MiddleWareProtocol
from unfazed.types import ASGIApp, Receive, Scope, Send

from unfazed.http import Request


class BaseMiddleWare(MiddleWareProtocol):
    def __init__(self, app: ASGIApp) -> None:
        self.app = app

    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> ASGIApp:
        if scope["type"] == "http":
            return await self.app(scope, receive, send)

        await self.app(scope, receive, send)



import typing as t
from importlib import import_module


def import_string(import_name: str) -> t.Type:
    """
    Import a class or object from a string representing the module path and object name.

    Args:
        import_name (str): The string representing the module path and object name.

    Returns:
        Type: The imported class or object.

    Raises:
        ImportError: If the import_name doesn't look like a valid module path.
    """
    try:
        module_name, class_name = import_name.rsplit(".", 1)
    except ValueError:
        raise ImportError(f"{import_name} doesn't look like a module path")

    module = import_module(module_name)
    return getattr(module, class_name)



import typing as t

from unfazed.protocol import Application, DataBaseProtocol


class ConnectionHandler:
    def __init__(self):
        self._connections: dict[str, DataBaseProtocol] = {}

        if t.TYPE_CHECKING:
            self._app = Application()

    def set_app(self, app: Application) -> None:
        self._app = app

    def get_app(self) -> Application:
        return self._app

    def set_connection(self, alias: str, connection: DataBaseProtocol):
        if alias in self._connections:
            raise ValueError(f"Connection {alias} already exists")

        self._connections[alias] = connection

    def get_connection(self, alias: str) -> DataBaseProtocol:
        if alias not in self._connections:
            raise ValueError(f"Connection {alias} does not exist")

        return self._connections[alias]

    def clear(self):
        self._connections.clear()

    def reset_connection(self, alias: str):
        if alias not in self._connections:
            raise ValueError(f"Connection {alias} does not exist")

        self._connections[alias].close()
        self._connections[alias].connect()


connections = ConnectionHandler()



import typing as t

from .orm import OrmModel


class AppCommand(t.Protocol):
    name: str
    import_path: str

    def execute(self) -> None: ...


class AppConfig(t.Protocol):
    name: str
    commands: t.Sequence[AppCommand]
    models: t.Sequence[t.Type[OrmModel]]

    def ready(self) -> None: ...



import os
import stat
import typing as t
import warnings
from email.utils import formatdate
from functools import partial
from mimetypes import guess_type
from urllib.parse import quote

import anyio
import anyio.to_thread
import orjson as json
from starlette._compat import md5_hexdigest
from starlette.background import BackgroundTask
from starlette.concurrency import iterate_in_threadpool
from starlette.datastructures import URL
from starlette.responses import Response as StarletteResponse

from unfazed.types import Receive, Scope, Send


class Response(StarletteResponse):
    pass


class JSONResponse(Response):
    media_type = "application/json"

    def render(self, content: t.Any) -> bytes:
        return json.dumps(content)


class PlainTextResponse(Response):
    media_type = "text/plain"


class RedirectResponse(Response):
    def __init__(
        self,
        url: str | URL,
        status_code: int = 307,
        headers: t.Mapping[str, str] | None = None,
        background: BackgroundTask | None = None,
    ) -> None:
        super().__init__(
            content=b"", status_code=status_code, headers=headers, background=background
        )
        self.headers["location"] = quote(str(url), safe=":/%#?=@[]!$&'()*+,;")


Content = t.Union[str, bytes]
SyncContentStream = t.Iterable[Content]
AsyncContentStream = t.AsyncIterable[Content]
ContentStream = t.Union[AsyncContentStream, SyncContentStream]


class StreamingResponse(Response):
    body_iterator: AsyncContentStream

    def __init__(
        self,
        content: ContentStream,
        status_code: int = 200,
        headers: t.Mapping[str, str] | None = None,
        media_type: str | None = None,
        background: BackgroundTask | None = None,
    ) -> None:
        if isinstance(content, t.AsyncIterable):
            self.body_iterator = content
        else:
            self.body_iterator = iterate_in_threadpool(content)
        self.status_code = status_code
        self.media_type = self.media_type if media_type is None else media_type
        self.background = background
        self.init_headers(headers)

    async def listen_for_disconnect(self, receive: Receive) -> None:
        while True:
            message = await receive()
            if message["type"] == "http.disconnect":
                break

    async def stream_response(self, send: Send) -> None:
        await send(
            {
                "type": "http.response.start",
                "status": self.status_code,
                "headers": self.raw_headers,
            }
        )
        async for chunk in self.body_iterator:
            if not isinstance(chunk, bytes):
                chunk = chunk.encode(self.charset)
            await send({"type": "http.response.body", "body": chunk, "more_body": True})

        await send({"type": "http.response.body", "body": b"", "more_body": False})

    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:
        async with anyio.create_task_group() as task_group:

            async def wrap(func: t.Callable[[], t.Awaitable[None]]) -> None:
                await func()
                task_group.cancel_scope.cancel()

            task_group.start_soon(wrap, partial(self.stream_response, send))
            await wrap(partial(self.listen_for_disconnect, receive))

        if self.background is not None:
            await self.background()


class FileResponse(Response):
    chunk_size = 64 * 1024

    def __init__(
        self,
        path: str | os.PathLike[str],
        status_code: int = 200,
        headers: t.Mapping[str, str] | None = None,
        media_type: str | None = None,
        background: BackgroundTask | None = None,
        filename: str | None = None,
        stat_result: os.stat_result | None = None,
        method: str | None = None,
        content_disposition_type: str = "attachment",
    ) -> None:
        self.path = path
        self.status_code = status_code
        self.filename = filename
        if method is not None:
            warnings.warn(
                "The 'method' parameter is not used, and it will be removed.",
                DeprecationWarning,
            )
        if media_type is None:
            media_type = guess_type(filename or path)[0] or "text/plain"
        self.media_type = media_type
        self.background = background
        self.init_headers(headers)
        if self.filename is not None:
            content_disposition_filename = quote(self.filename)
            if content_disposition_filename != self.filename:
                content_disposition = "{}; filename*=utf-8''{}".format(
                    content_disposition_type, content_disposition_filename
                )
            else:
                content_disposition = '{}; filename="{}"'.format(
                    content_disposition_type, self.filename
                )
            self.headers.setdefault("content-disposition", content_disposition)
        self.stat_result = stat_result
        if stat_result is not None:
            self.set_stat_headers(stat_result)

    def set_stat_headers(self, stat_result: os.stat_result) -> None:
        content_length = str(stat_result.st_size)
        last_modified = formatdate(stat_result.st_mtime, usegmt=True)
        etag_base = str(stat_result.st_mtime) + "-" + str(stat_result.st_size)
        etag = f'"{md5_hexdigest(etag_base.encode(), usedforsecurity=False)}"'

        self.headers.setdefault("content-length", content_length)
        self.headers.setdefault("last-modified", last_modified)
        self.headers.setdefault("etag", etag)

    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:
        if self.stat_result is None:
            try:
                stat_result = await anyio.to_thread.run_sync(os.stat, self.path)
                self.set_stat_headers(stat_result)
            except FileNotFoundError:
                raise RuntimeError(f"File at path {self.path} does not exist.")
            else:
                mode = stat_result.st_mode
                if not stat.S_ISREG(mode):
                    raise RuntimeError(f"File at path {self.path} is not a file.")
        await send(
            {
                "type": "http.response.start",
                "status": self.status_code,
                "headers": self.raw_headers,
            }
        )
        if scope["method"].upper() == "HEAD":
            await send({"type": "http.response.body", "body": b"", "more_body": False})
        elif "extensions" in scope and "http.response.pathsend" in scope["extensions"]:
            await send({"type": "http.response.pathsend", "path": str(self.path)})
        else:
            async with await anyio.open_file(self.path, mode="rb") as file:
                more_body = True
                while more_body:
                    chunk = await file.read(self.chunk_size)
                    more_body = len(chunk) == self.chunk_size
                    await send(
                        {
                            "type": "http.response.body",
                            "body": chunk,
                            "more_body": more_body,
                        }
                    )
        if self.background is not None:
            await self.background()



import orjson as json
from starlette.requests import Request as StarletteRequest


class Request(StarletteRequest):
    async def json(self):
        # replace the default json method with orjson
        if not hasattr(self, "_json"):
            body = await self.body()
            self._json = json.loads(body)
        return self._json



import typing as t
from threading import Lock

from starlette.middleware import Middleware

from unfazed.utils.module_loading import import_string


class MiddleWareCenter:
    def __init__(self, middleware_paths: list[str] = None) -> None:
        self._middleware_paths = middleware_paths or []
        self._ready = False
        self._lock = Lock()
        self.middlewares: t.Sequence[Middleware] = []

    def setup(self) -> None:
        if self._ready:
            return

        with self._lock:
            if self._ready:
                return
            for middleware_path in self._middleware_paths:
                middleware_cls = import_string(middleware_path)
                self.middlewares.append(middleware_cls)

        return None



import typing as t
from pathlib import Path

import unfazed
from click import Group
from unfazed.utils.module_loading import import_string

from .base import BaseCommand

if t.TYPE_CHECKING:
    from unfazed.core import Unfazed


class CommandCenter(Group):
    def __init__(
        self,
        unfazed: "Unfazed",
        name: str | None = None,
        commands: t.List[BaseCommand] | None = None,
    ) -> None:
        if commands is None:
            commands = []
        super().__init__(name=name, commands={c.name: c for c in commands})
        self.unfazed = unfazed
        self._ready = False

    def add_base_commands(self, include: t.Sequence[str] | None = None) -> None:
        enable_include = include is not None
        include = include or []
        internal_command_dir = Path(unfazed.__path__[0] + "/command/internal")
        for command_file in internal_command_dir.glob("*.py"):
            command_name = command_file.stem
            if command_name == "__init__":
                continue
            if enable_include and command_name not in include:
                continue
            command_kls: BaseCommand = import_string(
                f"unfazed.command.internal.{command_name}.Command"
            )

            escaped = command_name.replace("_", "-")
            cmd = command_kls(
                unfazed=self.unfazed,
                app_label="unfazed.command.internal",
                name=escaped,
            )
            self.add_command(cmd)

    def setup(self):
        if self._ready:
            return
        self.add_base_commands()

        self._ready = True
        return None

    def setup_cli(self):
        self.add_base_commands(include=["startproject"])



import os
import typing as t

_ = t.Optional

if t.TYPE_CHECKING:
    from unfazed.core import Unfazed

PathLike = t.Union[str, bytes, os.PathLike]
Context = t.Mapping[str, str | int | float]

type CanBeImported = t.Annotated[
    str,
    "split by dot, can be imported by importlib.import_module",
]
type Email = str
type TimeZone = str


type InstalledApp = t.Sequence[CanBeImported]
type Middlewares = t.Sequence[CanBeImported]


# TODO
# impl websockets protocol


class _Asgi(t.TypedDict):
    version: str
    spec_version: str


class _Scope(t.TypedDict):
    type: t.Annotated[str, "http"]
    asgi: _Asgi
    http_version: t.Annotated[
        str,
        "Version of the ASGI HTTP spec this server understands; one of '2.0', '2.1', '2.2' or '2.3'. Optional; if missing assume 2.0",
    ]
    method: str
    scheme: str
    path: str
    raw_path: bytes
    query_string: bytes
    root_path: str
    headers: t.Sequence[t.Tuple[bytes, bytes]]
    client: t.Tuple[str, int]
    server: t.Tuple[str, int]
    state: dict | None
    app: "Unfazed"


type Scope = _Scope


class _BodyEvent(t.TypedDict):
    type: str
    body: bytes | None
    more_body: bool | None


type ReceiveEvent = _BodyEvent

type Receive = t.Callable[[], t.Awaitable[ReceiveEvent]]


class _StartSendEvent(t.TypedDict):
    type: str
    status: int
    headers: t.Sequence[t.Tuple[bytes, bytes]]
    trailers: bool | None


type SendEvent = _StartSendEvent | _BodyEvent


type Send = t.Callable[[SendEvent], t.Awaitable[None]]

type ASGIApp = t.Callable[[Scope, Receive, Send], t.Awaitable[None]]



import typing as t

from pydantic import BaseModel

from .base import DataBaseWapper
from ..settings import BaseSettings as _BaseSettings


class Options(BaseModel):
    min_size: int | None = None
    max_size: int | None = None
    pool_recycle: int | None = None
    ssl: bool | None = None


class BaseSettings(_BaseSettings):
    OPTIONS: Options = Options()


class MysqlWrapper(DataBaseWapper):
    def __init__(self, settings_dict: dict[str, t.Any]):
        self.config = BaseSettings(**settings_dict)



from typing import Any

from databases import Database
from databases.core import DatabaseURL


class DataBaseWapper(Database):
    def __init__(
        self, url: str | DatabaseURL, *, force_rollback: bool = False, **options: Any
    ):
        super().__init__(url, force_rollback=force_rollback, **options)



import asyncio
import typing as t
from pathlib import Path

from click import Command as ClickCommand
from click import Option, Parameter
from jinja2 import Template
from unfazed import types as tp
from unfazed.protocol import BaseCommandMethod

if t.TYPE_CHECKING:
    from unfazed.core import Unfazed


class BaseCommand(ClickCommand, BaseCommandMethod):
    """
    Usage:
        class MyCommand(BaseCommand):
            help_text = "This is a help text"

        def add_arguments(self) -> List[Parameter]:
            return [
                Option(
                    ["--name", "-n"],
                    type=str,
                    help="The name of the project",
                ),
                Option(
                    ["--directory"],
                    type=str,
                    help="The directory of the project",
                ),
            ]

        async def handle(self, name: str, directory: str) -> None:
            await asyncio.sleep(1)
            print("MyCommand.handle name = {}, dir = {}".format(name, directory))


    """

    help_text = ""

    def __init__(
        self,
        unfazed: "Unfazed",
        name: str,
        app_label: str,
        *,
        context_settings: t.MutableMapping[str, t.Any] | None = None,
        callback: t.Callable[..., t.Any] | None = None,
        params: t.List[Parameter] | None = None,
        help: str | None = None,
        epilog: str | None = None,
        short_help: str | None = None,
        options_metavar: str | None = "[OPTIONS]",
        add_help_option: bool = True,
        no_args_is_help: bool = False,
        hidden: bool = False,
        deprecated: bool = False,
    ) -> None:
        help = help or self.help_text
        params = params or []
        params.extend(self.add_arguments())
        callback = self._callback

        self.unfazed = unfazed
        self.app_label = app_label

        super().__init__(
            name,
            context_settings,
            callback,
            params,
            help,
            epilog,
            short_help,
            options_metavar,
            add_help_option,
            no_args_is_help,
            hidden,
            deprecated,
        )

    def _callback(self, **option: t.Optional[t.Any]) -> None:
        if not asyncio.iscoroutinefunction(self.handle):
            raise TypeError("handle method must be a coroutine")
        asyncio.run(self.handle(**option))


class BaseCommandWithTemplate(BaseCommand):
    def add_arguments(self) -> t.Sequence[Parameter]:
        cwd = Path.cwd()
        return [
            Option(
                ["--name", "-n"],
                type=str,
                help="The name of the project",
            ),
            Option(
                ["--directory", "-d"],
                type=str,
                default=str(cwd),
                help="The directory where the project will be created",
                show_default=True,
            ),
            Option(
                ["--template", "-t"],
                type=str,
                help="The template use to render the project",
                default=self.template,
                show_default=True,
            ),
        ]

    async def handle(
        self,
        name: str,
        directory: tp.PathLike,
        template: tp.PathLike,
    ) -> None:
        render_template = template

        if isinstance(render_template, str):
            render_template = Path(render_template)
        render_dest = Path(directory)

        if not render_dest.exists():
            render_dest.mkdir(parents=True, exist_ok=True)

        context = {
            "project": name,
            "app_name": name,
            "version": self.unfazed.version,
        }

        for root, _, files in render_template.walk():
            for f in files:
                new_root = render_dest / root.relative_to(render_template)
                self.handle_file(root, new_root, f, context)

    def handle_file(
        self,
        root: Path,
        new_root: Path,
        file_path: tp.PathLike,
        context: tp.Context,
    ) -> None:
        new_root_rendered = Path(Template(str(new_root)).render(**context))
        if not new_root_rendered.exists():
            new_root_rendered.mkdir(parents=True, exist_ok=True)

        cur_file = root.joinpath(file_path)
        new_file = new_root_rendered.joinpath(file_path)
        with open(cur_file, "r") as cur_fh, open(new_file, "w") as new_fh:
            rendered_text = Template(cur_fh.read(), autoescape=True).render(**context)
            new_fh.write(rendered_text)



import typing as t

from pydantic import BaseModel, ConfigDict

from unfazed.protocol import OrmMeta

# type hinting
_ = t.Optional
M = t.TypeVar("M", bound="Model")
Ins = t.TypeVar("Ins", bound="BaseModel")


class Meta(OrmMeta):
    def __init__(
        self,
        db_table: _[str] = None,
        abstract: bool = False,
        app_label: _[str] = None,
        db_table_comment: _[str] = "",
        default_order_by: _[list[str]] = None,
        indexes: _[list[str]] = None,
        unique_together: _[list[str]] = None,
    ):
        self._db_table = db_table
        self.abstract = abstract
        self._app_label = app_label
        self.db_table_comment = db_table_comment
        self.default_order_by = default_order_by
        self.indexes = indexes
        self.unique_together = unique_together

        self._database = None
        self._model: M = None

    def set_database(self, database):
        self._database = database

    def set_model(self, model: t.Type[M]):
        self._model = model

    @property
    def db_table(self):
        if not self._db_table:
            self._db_table = f"{self.app_label}_{self._model.__name__.lower()}"
        return self._db_table

    @property
    def app_label(self):
        if not self._app_label:
            self._app_label = self._model.__module__.split(".")[-2]
        return self._app_label

    def build_table(self):
        pass


class Model(BaseModel):
    """

    Usage:

    ```python

    from unfazed_orm.models import Model, Meta

    class User(Model):
        meta = Meta(db_table="users")

    """

    meta: Meta
    # objects = QuerySet()
    model_config = ConfigDict(arbitrary_types_allowed=True)

    async def save(self):
        return self

    async def delete(self):
        return self

    async def load(self):
        return self



import typing as t

_ = t.Optional


class OrmMeta(t.Protocol):
    db_table: _[str]
    abstract: bool
    app_label: _[str]
    db_table_comment: _[str]
    default_order_by: _[list[str]]
    indexes: _[list[str]]
    unique_together: _[list[str]]

    def set_database(self, database) -> None: ...
    def set_model(self, model: t.Type["OrmModel"]) -> None: ...


class OrmQuerySet[Ins](t.Protocol):
    _using: str

    def __init__(self) -> None: ...
    def using(self, alias: str) -> t.Self: ...
    def filter(self, **kwargs) -> t.Self: ...
    def exclude(self, **kwargs) -> t.Self: ...
    async def all(self) -> t.List[Ins]: ...
    async def get(self, **kwargs) -> Ins: ...
    async def create(self, **kwargs) -> Ins: ...
    async def update(self, **kwargs) -> None: ...
    async def delete(self, **kwargs) -> None: ...
    async def bulk_create(self, **kwargs) -> None: ...
    async def get_or_create(self, **kwargs) -> t.Tuple[bool, Ins]: ...
    async def update_or_create(self, **kwargs) -> t.Tuple[bool, Ins]: ...


class OrmModel[Ins](t.Protocol):
    objects: OrmQuerySet
    meta: OrmMeta

    async def save(self) -> Ins: ...
    async def delete(self) -> None: ...
    async def load(self) -> Ins: ...


class OrmDataBase(t.Protocol):
    def __init__(self, *args, **kwargs) -> None: ...
    async def connect(self) -> None: ...
    async def disconnect(self) -> None: ...
    async def __aenter__(self) -> t.Self: ...
    async def __aexit__(self, *args: t.Any, **kwargs: t.Any) -> None: ...
    async def fetch_all(
        self, query: t.Any, values: t.Optional[dict] = None
    ) -> t.List[t.Any]: ...
    async def fetch_one(
        self, query: t.Any, values: t.Optional[dict] = None
    ) -> t.Optional[t.Any]: ...
    async def fetch_val(
        self, query: t.Any, values: t.Optional[dict] = None, column: t.Any = 0
    ) -> t.Any: ...
    async def execute(self, query: t.Any, values: t.Optional[dict] = None) -> t.Any: ...
    async def execute_many(self, query: t.Any, values: list) -> None: ...
    async def iterate(
        self, query: t.Any, values: t.Optional[dict] = None
    ) -> t.AsyncGenerator[t.Mapping, None]: ...
    def connection(self) -> t.Self: ...
    async def reset(self) -> None: ...
